#include "SimpleSynthNode.h"
#include "SimpleSynthAddOn.h"

#include <stdio.h>
#include <string.h>
#include <Window.h>
#include <View.h>
#include <MediaTheme.h>

#define DEBUG 1
#if DEBUG
	#define PRINT printf
#else
	static inline void PRINT( const char*, ...) {}
#endif

#if B_HOST_IS_BENDIAN
	#define ENDIAN B_MEDIA_BIG_ENDIAN
#else
	#define ENDIAN B_MEDIA_LITTLE_ENDIAN
#endif

#define FREQ 440.0

//////////////////////////////////////////
// Constructor/Desctructor
SimpleSynthNode::SimpleSynthNode( BMediaAddOn* addon )
:	BMediaNode( "SimpleSynthNode" ),
	BBufferProducer( B_MEDIA_RAW_AUDIO ),
	BControllable(),
	BMediaEventLooper(),

	mAddOn( addon ),
	
	mBufferGroup(NULL),

	mLatency(0),
	mInternalLatency(0),
	mOutputEnabled(true),
	mStartTime(0),
	mFramesSent(0),

	mGain(0.1),
	mGainLastChanged(0),
	mMidiChannel(1),
	mMidiChannelChanged(0),
	mPulseWidth(50.0),
	mPulseWidthChanged(0),
	mLastPhase(0.0)
{
	PRINT("SimpleSynthNode::SimpleSynthNode()\n");

	mPreferFormat.type = B_MEDIA_RAW_AUDIO;
	mPreferFormat.u.raw_audio.format = media_raw_audio_format::B_AUDIO_FLOAT;
	mPreferFormat.u.raw_audio.channel_count = 2;
	mPreferFormat.u.raw_audio.frame_rate = 44100;
	mPreferFormat.u.raw_audio.byte_order = ENDIAN;
	mPreferFormat.u.raw_audio.buffer_size = media_raw_audio_format::wildcard.buffer_size;
}

SimpleSynthNode::
~SimpleSynthNode()
{
	PRINT("SimpleSynthNode::~SimpleSynthNode()\n");
	Quit();
}

//////////////////////////////////////////
// BMediaNode methods

BMediaAddOn*
SimpleSynthNode::
AddOn( int32 *internal_id ) const
{
	PRINT("SimpleSynthNode::AddOn()\n");

	if ( mAddOn )
		*internal_id = 0;
		
	return mAddOn;
}


void 
SimpleSynthNode::
NodeRegistered()
{
	int i;
	char buf[16];

	BMediaEventLooper::NodeRegistered();

	BParameterWeb* web = new BParameterWeb();
	BParameterGroup* mainGroup = web->MakeGroup( "SympleSynth Parameter Group" );

	BParameterGroup* group = mainGroup->MakeGroup( "Gain group" );
	group->MakeNullParameter( 
		GAIN_NULL_PARAM, B_MEDIA_NO_TYPE, "Output", B_WEB_BUFFER_OUTPUT );
	group->MakeContinuousParameter( 
		GAIN_PARAM, B_MEDIA_NO_TYPE, "Gain", B_GAIN, "", 0.0, 1.0, 0.01 );

	group = mainGroup->MakeGroup( "Synth group" );	
	group->MakeContinuousParameter( 
		PULSE_WIDTH_PARAM, B_MEDIA_NO_TYPE, "Pulse Width", B_LEVEL, "%", 5, 95, 0.2 );

	BDiscreteParameter* discrete_param;
	group = mainGroup->MakeGroup( "Midi group" );
	discrete_param = group->MakeDiscreteParameter( 
		MIDI_CHANNEL_PARAM, B_MEDIA_NO_TYPE, "MIDI Channel", B_TRACK );
	for ( i=1; i<=16; i++ ) {
		sprintf( buf, "%d", i );
		discrete_param->AddItem( i, buf );
	}
	
	SetParameterWeb( web );

	mOutput.node = Node();
	mOutput.source.port = ControlPort();
	mOutput.source.id = mOutput.node.port;
	mOutput.destination = media_destination::null;
	mOutput.format = mPreferFormat;
	strcpy( mOutput.name, Name() );
}

//////////////////////////////////////////
// BBufferProducer methods

status_t 
SimpleSynthNode::
FormatSuggestionRequested(
		media_type in_type, 
		int32 in_quality, 
		media_format* out_format )
{
	PRINT("SimpleSynthNode::FormatSuggestionRequested()\n");
	
	if ( !out_format )
		return B_BAD_VALUE;

	if ( in_type != B_MEDIA_RAW_AUDIO && in_type != B_MEDIA_UNKNOWN_TYPE )
		return B_MEDIA_BAD_FORMAT;

	*out_format = mPreferFormat;
	
	return B_OK;
}

status_t 
SimpleSynthNode::
FormatProposal(
		const media_source& in_source, 
		media_format* io_format )
{
	PRINT("SimpleSynthNode::FormatProposal()\n");

	if ( in_source != mOutput.source ) {
		PRINT("\treturning B_MEDIA_BAD_SOURCE\n");
		return B_MEDIA_BAD_SOURCE;
	}

	media_type requestedType = io_format->type;
	*io_format = mPreferFormat;

	if ( requestedType != B_MEDIA_UNKNOWN_TYPE && requestedType != B_MEDIA_RAW_AUDIO ) {
		PRINT("\treturning B_MEDIA_BAD_FORMAT\n");
		return B_MEDIA_BAD_FORMAT;
	}
	
	return B_OK;
}

status_t 
SimpleSynthNode::
FormatChangeRequested(
		const media_source& in_source, 
		const media_destination& in_destination, 
		media_format* io_format, 
		int32* /*_deprecated_*/ )
{
	PRINT("SimpleSynthNode::FormatChangeRequested()\n");
	return B_ERROR;
}

status_t 
SimpleSynthNode::
GetNextOutput(
		int32* cookie, 
		media_output* out_output )
{
	PRINT("SimpleSynthNode::GetNextOutput()\n");

	if (0 == *cookie) {
		*out_output = mOutput;
		*cookie += 1;
		return B_OK;
	}
	
	return B_BAD_INDEX;
}

status_t 
SimpleSynthNode::
DisposeOutputCookie( int32 cookie )
{
	PRINT("SimpleSynthNode::DisposeOutputCookie()\n");
	return B_OK;
}

status_t 
SimpleSynthNode::
SetBufferGroup(
		const media_source& in_source, 
		BBufferGroup* newGroup )
{
	PRINT("SimpleSynthNode::SetBufferGroup()\n");

	if ( in_source != mOutput.source )
		return B_MEDIA_BAD_SOURCE;

	if ( newGroup == mBufferGroup )
		return B_OK;

	delete mBufferGroup;

	if ( newGroup != NULL )
		mBufferGroup = newGroup;
	else
		AllocateBuffers();

	return B_OK;
}

status_t 
SimpleSynthNode::
PrepareToConnect(
		const media_source& in_source,
		const media_destination& in_destination,
		media_format* io_format,
		media_source* out_source,
		char* out_name )
{
	PRINT("SimpleSynthNode::PrepareToConnect()\n");

	if ( in_source != mOutput.source )
		return B_MEDIA_BAD_SOURCE;

	if ( mOutput.destination != media_destination::null )
		return B_MEDIA_ALREADY_CONNECTED;

	if ( io_format->type != B_MEDIA_RAW_AUDIO ) {
		PRINT( "\tnon-raw-audio format\n" );
		return B_MEDIA_BAD_FORMAT;
	} else if ( io_format->u.raw_audio.format != media_raw_audio_format::B_AUDIO_FLOAT ) {
		PRINT( "\tnon-float-audio format\n" );
		return B_MEDIA_BAD_FORMAT;
	}

	if ( io_format->u.raw_audio.buffer_size == media_raw_audio_format::wildcard.buffer_size ) {
		io_format->u.raw_audio.buffer_size = 2048;
	}

	mOutput.destination = in_destination;
	mOutput.format = *io_format;
	*out_source = mOutput.source;
	strncpy( out_name, mOutput.name, B_MEDIA_NAME_LENGTH );

	return B_OK;
}

void 
SimpleSynthNode::
Connect(
		status_t error, 
		const media_source& in_source,
		const media_destination& in_destination,
		const media_format& in_format,
		char* io_name )
{
	PRINT("SimpleSynthNode::Connect()\n");

	if ( error ) {
		mOutput.destination = media_destination::null;
		mOutput.format = mPreferFormat;
		return;
	}

	mOutput.destination = in_destination;
	mOutput.format = in_format;
	strncpy( io_name, mOutput.name, B_MEDIA_NAME_LENGTH );

	media_node_id timeSource;
	FindLatencyFor( mOutput.destination, &mLatency, &timeSource );
	PRINT( "\tdownstream latency = %Ldusec\n", mLatency );

	bigtime_t start, end;
	float* data = (float*)malloc( mOutput.format.u.raw_audio.buffer_size );
	start = ::system_time();
	FillBuffer( data );
	end = ::system_time();
	mInternalLatency = end - start;
	free(data);
	PRINT("\tbuffer-filling took %Ld usec on this machine\n", mInternalLatency);

	SetEventLatency( mLatency + mInternalLatency );

	size_t framesPerBuffer = 
		mOutput.format.u.raw_audio.buffer_size / sizeof(float) 
		/ mOutput.format.u.raw_audio.channel_count;
	bigtime_t duration = 
		bigtime_t(1000000) * framesPerBuffer 
		/ bigtime_t(mOutput.format.u.raw_audio.frame_rate);

	SetBufferDuration( duration );

	PRINT("\tbufferlength: %ld\n\tframesPerBuffer: %ld\n",
			in_format.u.raw_audio.buffer_size, framesPerBuffer );

	if (!mBufferGroup)
		AllocateBuffers();
}

void 
SimpleSynthNode::
Disconnect(
		const media_source& in_source,
		const media_destination& in_destination )
{
	if ( in_destination == mOutput.destination 
			&& in_source == mOutput.source ) {
		mOutput.destination = media_destination::null;
		mOutput.format = mPreferFormat;
		delete mBufferGroup;
		mBufferGroup = NULL;
	}
		
	PRINT("SimpleSynthNode::Disconnect()\n");
}

void 
SimpleSynthNode::
LateNoticeReceived(
		const media_source& in_source,
		bigtime_t how_much,
		bigtime_t performance_time )
{
	PRINT("SimpleSynthNode::LateNoticeReceived()\n");

	if ( RunMode() == B_RECORDING )
		PRINT( "\tdo nothing as in recording mode\n" );
	else if ( RunMode() == B_INCREASE_LATENCY ) {
		mInternalLatency += how_much;
		SetEventLatency(mLatency + mInternalLatency);
		PRINT( "\tincreasing latency to %Ld\n", mLatency + mInternalLatency);
	} else {
		size_t nFrames = mOutput.format.u.raw_audio.buffer_size / sizeof(float) 
				/ mOutput.format.u.raw_audio.channel_count;
		mFramesSent += nFrames;
		PRINT( "\tskipping a buffer to try to catch up\n");
	}
}

void 
SimpleSynthNode::
EnableOutput(
		const media_source& in_source,
		bool enabled,
		int32* /*_deprecated_*/ )
{
	PRINT("SimpleSynthNode::EnableOutput()\n");

	mOutputEnabled = enabled;
}

status_t 
SimpleSynthNode::
HandleMessage( 
	int32 message, 
	const void *data, 
	size_t size )
{
	PRINT("SimpleSynthNode::HandleMessage()\n");

	if ( BBufferProducer::HandleMessage( message, data, size ) != B_OK 
			&& BMediaEventLooper::HandleMessage( message, data, size ) != B_OK 
			&& BControllable::HandleMessage( message, data, size ) != B_OK 
			&& BMediaNode::HandleMessage( message, data, size ) != B_OK ) {
		 BMediaNode::HandleBadMessage( message, data, size );
		 return B_ERROR;
	}

	return B_OK;
}

//////////////////////////////////////////
// BControllable methods

status_t 
SimpleSynthNode::
GetParameterValue(
		int32 parameterID, 
		bigtime_t *lastChangeTime, 
		void *value, 
		size_t *ioSize)
{
	PRINT("SimpleSynthNode::GetParameterValue()\n");
	
	if ( !lastChangeTime || !value || !ioSize ) {
		PRINT("\tNull pointer.()\n");
		return B_ERROR;
	}
	
	switch( parameterID ) {

	case GAIN_PARAM:
		if ( *ioSize < sizeof(float) ) {
			PRINT("\tValue buffer insufficient.\n");
			return B_ERROR;
		}	
		*(float*)value = mGain;
		*ioSize = sizeof(float);
		*lastChangeTime = mGainLastChanged;
		break;

	case MIDI_CHANNEL_PARAM:
		if ( *ioSize < sizeof(int32) ) {
			PRINT("\tValue buffer insufficient.\n");
			return B_ERROR;
		}		
		*(int32*)value = mMidiChannel;
		*ioSize = sizeof(int32);
		*lastChangeTime = mMidiChannelChanged;
		break;
	
	case PULSE_WIDTH_PARAM:
		if ( *ioSize < sizeof(float) ) {
			PRINT("\tValue buffer insufficient.\n");
			return B_ERROR;
		}		
		*(float*)value = mPulseWidth;
		*ioSize = sizeof(float);
		*lastChangeTime = mPulseWidthChanged;
		break;
	
	default:
		PRINT("\tGot an unknown parameter ID.\n");
		return B_ERROR;

	}
	
	return B_OK;
}

void 
SimpleSynthNode::
SetParameterValue(
		int32 parameterID, 
		bigtime_t changeTime,
		const void *value, 
		size_t size)
{
	PRINT("SimpleSynthNode::SetParameterValue()\n");

	if ( !value ) {
		PRINT("\tNull pointer\n");
		return;
	}

	switch( parameterID ) {

	case GAIN_PARAM:
		if ( size < sizeof(float) ) {
			PRINT("\tValue buffer insufficient.\n");
			return;
		} else {
			media_timed_event event( changeTime, BTimedEventQueue::B_PARAMETER );
			event.data = sizeof(float);
			event.bigdata = GAIN_PARAM;
			*(float*)event.user_data = *(float*)value;
			EventQueue()->AddEvent( event );
		}
		break;

	case MIDI_CHANNEL_PARAM:
		if ( size < sizeof(int32) ) {
			PRINT("\tValue buffer insufficient.\n");
			return;
		} else {
			media_timed_event event( changeTime, BTimedEventQueue::B_PARAMETER );
			event.data = sizeof(int32);
			event.bigdata = MIDI_CHANNEL_PARAM;
			*(int32*)event.user_data = *(int32*)value;
			EventQueue()->AddEvent( event );
		}
		break;

	case PULSE_WIDTH_PARAM:
		if ( size < sizeof(float) ) {
			PRINT("\tValue buffer insufficient.\n");
			return;
		} else {
			media_timed_event event( changeTime, BTimedEventQueue::B_PARAMETER );
			event.data = sizeof(float);
			event.bigdata = PULSE_WIDTH_PARAM;
			*(float*)event.user_data = *(float*)value;
			EventQueue()->AddEvent( event );
		}
		break;

	default:	
		break;
	}
}

//////////////////////////////////////////
// BMediaEventLooper methods

void 
SimpleSynthNode::
HandleEvent(
		const media_timed_event* event,
		bigtime_t lateness,
		bool realTimeEvent )
{
//	PRINT("SimpleSynthNode::HandleEvent()\n");
	size_t dataSize;
	int32 param;
	float newValueFloat;
	int32 newValueInt32;
	
	switch (event->type) {

	case BTimedEventQueue::B_START:
		PRINT("Handling B_START event\n");

		if (RunState() != B_STARTED) {
			mFramesSent = 0;
			mStartTime = event->event_time;
			media_timed_event firstBufferEvent( mStartTime, BTimedEventQueue::B_HANDLE_BUFFER );
			EventQueue()->AddEvent(firstBufferEvent);
		}
		break;

	case BTimedEventQueue::B_STOP:
		PRINT("Handling B_STOP event\n");

		EventQueue()->FlushEvents(0, BTimedEventQueue::B_ALWAYS, true, BTimedEventQueue::B_HANDLE_BUFFER);
		break;

	case BTimedEventQueue::B_PARAMETER:
		PRINT("Handling B_PARAMETER event\n");
		
		param = int32(event->bigdata);
		dataSize = size_t(event->data);
		switch (param) {
		case GAIN_PARAM:
			PRINT("\tHandling GAIN_PARAM\n");
			
			if (dataSize < sizeof(float)) {
				PRINT("\tShort of dataSize\n");
				break;
			}

			newValueFloat = *(float*)event->user_data;
			if (mGain != newValueFloat) {
				mGain = newValueFloat;
				mGainLastChanged = TimeSource()->Now();
				BroadcastNewParameterValue(
					mGainLastChanged, param, &mGain, sizeof(float));
			}
			break;

		case MIDI_CHANNEL_PARAM:
			PRINT("\tHandling MIDI_CHANNEL_PARAM\n");

			if (dataSize < sizeof(int32)) {
				PRINT("\tShort of dataSize\n");
				break;
			}
			
			newValueInt32 = *(int32*)event->user_data;
			if (mMidiChannel != newValueInt32) {
				mMidiChannel = newValueInt32;
				mMidiChannelChanged = TimeSource()->Now();
				BroadcastNewParameterValue(
					mMidiChannelChanged, param, &mMidiChannel, sizeof(int32));
			}

			break;

		case PULSE_WIDTH_PARAM:
			PRINT("\tHandling PULSE_WIDTH_PARAM\n");
			
			if (dataSize < sizeof(float)) {
				PRINT("\tShort of dataSize\n");
				break;
			}

			newValueFloat = *(float*)event->user_data;
			if (mPulseWidth != newValueFloat) {
				mPulseWidth = newValueFloat;
				mPulseWidthChanged = TimeSource()->Now();
				BroadcastNewParameterValue(
					mPulseWidthChanged, param, &mPulseWidth, sizeof(float));
			}
			break;

		default:
			PRINT("Got a B_PARAMETER event for a parameter we don't have.(%ld)\n", param);
		}
		break;

	case BTimedEventQueue::B_HANDLE_BUFFER:
//		PRINT("Handling B_HANDLE_BUFFER event\n");
		if ( RunState() == BMediaEventLooper::B_STARTED 
			&& mOutput.destination != media_destination::null )
		{
			BBuffer* buffer = FillNextBuffer( event->event_time );
			if (buffer) {
				status_t err = B_ERROR;
				if ( mOutputEnabled )
					err = SendBuffer(buffer, mOutput.destination);
				if ( err )
					buffer->Recycle();
			}

			size_t nFrames = mOutput.format.u.raw_audio.buffer_size / sizeof(float) 
					/ mOutput.format.u.raw_audio.channel_count;
			mFramesSent += nFrames;

			bigtime_t nextEvent = mStartTime + bigtime_t( double(mFramesSent) 
					/ double(mOutput.format.u.raw_audio.frame_rate) * 1000000.0);
			media_timed_event nextBufferEvent( nextEvent, BTimedEventQueue::B_HANDLE_BUFFER );
			EventQueue()->AddEvent( nextBufferEvent );
		}
		break;

	default:
		break;
	}
}

//////////////////////////////////////////
// Heart of the producer

BBuffer* 
SimpleSynthNode::
FillNextBuffer( bigtime_t eventTime )
{
	BBuffer* buf = mBufferGroup->RequestBuffer(
		mOutput.format.u.raw_audio.buffer_size, BufferDuration());

	if (!buf) {
		PRINT("Failed to get buffer. at %Ld\n", eventTime );
		return NULL;
	}

	FillBuffer( (float*)buf->Data() );

	bigtime_t stamp;
	if (RunMode() == B_RECORDING) {
		stamp = eventTime;
	} else {
		stamp = mStartTime 
			+ bigtime_t( double(mFramesSent) 
			/ double(mOutput.format.u.raw_audio.frame_rate) * 1000000.0 );
	}
	media_header* hdr = buf->Header();
	hdr->start_time = stamp;
	hdr->type = B_MEDIA_RAW_AUDIO;
	hdr->size_used = mOutput.format.u.raw_audio.buffer_size;
	hdr->time_source = TimeSource()->ID();

	return buf;
}

void 
SimpleSynthNode::
FillBuffer( float* data )
{
	int32 samples = mOutput.format.u.raw_audio.buffer_size / sizeof(float);
	double one_cycle =  mOutput.format.u.raw_audio.frame_rate / FREQ;
	double half_cycle = one_cycle * mPulseWidth / 100.0;
	
	int32 i = 0;
	double j = mLastPhase;
	
	if ( j >= half_cycle ) {
		for ( ; j < one_cycle && i < samples; j += 1.0, i += 2 )
			data[i] = data[i+1] = -mGain;
		if ( j >= one_cycle )
			j -= one_cycle;
	}
	
	for ( ; i < samples; ) {
		for ( ; j < half_cycle && i < samples; j += 1.0, i += 2 )
			data[i] = data[i+1] = mGain;
		for ( ; j < one_cycle && i < samples; j += 1.0, i += 2 )
			data[i] = data[i+1] = -mGain;

		if ( j >= one_cycle )
			j -= one_cycle;
	}

	mLastPhase = j;
}

void 
SimpleSynthNode::
AllocateBuffers()
{
	PRINT("ToneProducer::AllocateBuffers\n");

	size_t size = mOutput.format.u.raw_audio.buffer_size;
	int32 count = int32( (mLatency + mInternalLatency) / BufferDuration() + 1 + 1 );
	mBufferGroup = new BBufferGroup(size, count);

	PRINT("\tlatency = %Ldusec, buffer duration = %Ldusec\n", 
			mLatency, BufferDuration());
	PRINT("\tcreating group of %ld buffers, size = %lu\n", count, size);
}

